// include the library code:
#include <LiquidCrystal.h>
#include <SPI.h>
#include <SD.h>

// macros for rows
#define ALIEN_ROW 0
#define SHIP_ROW 1

// macros for screen dimensions
#define HEIGHT 4
#define WIDTH 16

// macros for screen buffer
#define NO_DISPLAY 0
#define SHIP_DISPLAY 1
#define ALIEN_DISPLAY 2
#define BULLET_DISPLAY 3

// macros for shapes
#define SHIP 0
#define SHIP_BULLET 1
#define ALIEN 2
#define ALIEN_BULLET 3
#define UP_BULLET 4
#define DOWN_BULLET 5

// macros for menus
#define MENU_PLAY 0
#define MENU_LEADERBOARD 1
#define LDR_PG_1 2
#define LDR_PG_2 3

// macros for other game aspects
#define MAX_LEVEL 5
#define MAX_TIME 999
#define LEADERBOARD_ENTRIES 3

// initialize the LCD library by associating any needed LCD interface pin
// with the arduino pin number it is connected to
const int rs = 8, en = 9, d4 = 5, d5 = 4, d6 = 3, d7 = 2;
LiquidCrystal lcd(rs, en, d4, d5, d6, d7);

// initialize joystick pins
const int joystick_X = A1;
const int joystick_Y = A2;
const int joystick_BTN = 7;

// initialize buzzer pin
const int buzzer = 6;

// initialize SD module pin
const int SD_module_CS = 10;

// create shapes needed for the game
byte ship_shape[8] = {
  0b00000,
  0b00000,
  0b00000,
  0b00000,
  0b00000,
  0b00100,
  0b01110,
  0b11011
};

byte ship_bullet_shape[8] = {
  0b00000,
  0b00100,
  0b00100,
  0b00000,
  0b00000,
  0b00100,
  0b01110,
  0b11011
};

byte alien_shape[8] = {
  0b01010,
  0b10101,
  0b01110,
  0b10001,
  0b00000,
  0b00000,
  0b00000,
  0b00000
};

byte alien_bullet_shape[8] = {
  0b01010,
  0b10101,
  0b01110,
  0b10001,
  0b00000,
  0b00100,
  0b00100,
  0b00000
};

byte up_bullet_shape[8] = {
  0b00000,
  0b00100,
  0b00100,
  0b00000,
  0b00000,
  0b00000,
  0b00000,
  0b00000
};

byte down_bullet_shape[8] = {
  0b00000,
  0b00000,
  0b00000,
  0b00000,
  0b00000,
  0b00100,
  0b00100,
  0b00000
};

// global variables and structures
bool play_game = false;
int menu = MENU_PLAY;

bool see_leaderboard = false;
long int leaderboard[LEADERBOARD_ENTRIES];
int leaderboard_page = LDR_PG_1;
File myFile;

long int start_time;
long int end_time;

int screen[HEIGHT][WIDTH];

int notes[16] = {
  262,  // C4
  294,  // D4
  330,  // E4
  349,  // F4
  392,  // G4
  440,  // A4
  494,  // B4
  523,  // C5
  494,  // B4
  440,  // A4
  392,  // G4
  349,  // F4
  330,  // E4
  294,  // D4
  262,  // C4
  196   // G3
};

int victory_notes[16] = {
  523,  // C5
  659,  // E5
  784,  // G5
  1046, // C6
  784,  // G5
  659,  // E5
  523,  // C5
  392,  // G4
  440,  // A4
  523,  // C5
  659,  // E5
  784,  // G5
  523,  // C5
  659,  // E5
  784,  // G5
  1046  // C6
};

struct entity {
  int pos;
  bool bullet_active;
  int bullet_pos;
  int bullet_row;
};

entity ship = {0, false, -1, -1};

int level;
int level_movement_timer;
int level_direction_steps;

int alien_no;
int alien_movement_direction;
int alien_movement_timer;
int alien_direction_steps;

// game functions
void drawShape(int shape, int pos, int row) {
  lcd.setCursor(pos, row);
  lcd.write(byte(shape));
}

void updateScreen() {
  lcd.clear();
  int row;

  row = 0;
  for (int pos = 0; pos < WIDTH; pos++) {
    if (screen[row][pos] == ALIEN_DISPLAY && screen[row + 1][pos] == NO_DISPLAY)
      drawShape(ALIEN, pos, ALIEN_ROW);
    else if (screen[row][pos] == ALIEN_DISPLAY && screen[row + 1][pos] == BULLET_DISPLAY)
      drawShape(ALIEN_BULLET, pos, ALIEN_ROW);
    else if (screen[row][pos] == BULLET_DISPLAY && screen[row + 1][pos] == NO_DISPLAY)
      drawShape(UP_BULLET, pos, ALIEN_ROW);
    else if (screen[row][pos] == NO_DISPLAY && screen[row + 1][pos] == BULLET_DISPLAY)
      drawShape(DOWN_BULLET, pos, ALIEN_ROW);
  }

  row = 2;
  for (int pos = 0; pos < WIDTH; pos++) {
    if (screen[row][pos] == NO_DISPLAY && screen[row + 1][pos] == SHIP_DISPLAY)
      drawShape(SHIP, pos, SHIP_ROW);
    else if (screen[row][pos] == BULLET_DISPLAY && screen[row + 1][pos] == SHIP_DISPLAY)
      drawShape(SHIP_BULLET, pos, SHIP_ROW);
    else if (screen[row][pos] == BULLET_DISPLAY && screen[row + 1][pos] == NO_DISPLAY)
      drawShape(UP_BULLET, pos, SHIP_ROW);
    else if (screen[row][pos] == NO_DISPLAY && screen[row + 1][pos] == BULLET_DISPLAY)
      drawShape(DOWN_BULLET, pos, SHIP_ROW);
  }
}

void displayWelcomeScreen() {
  lcd.setCursor(0, 0);
  lcd.print("Arduino");
  lcd.setCursor(2, 1);
  lcd.print("Space Invaders");
  for (int positionCounter = 0; positionCounter < WIDTH; positionCounter++) {
    lcd.scrollDisplayRight();
  }
  for (int positionCounter = 0; positionCounter < WIDTH; positionCounter++) {
    tone(buzzer, notes[positionCounter]);
    lcd.scrollDisplayLeft();
    delay(150);
  }
  delay(500);
  noTone(buzzer);
  delay(1500);
  lcd.clear();
}

void displayLevelScreen(int level) {
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("Level ");
  lcd.print(level);
  for (int positionCounter = 0; positionCounter < WIDTH / 2; positionCounter++) {
    lcd.scrollDisplayLeft();
  }
  for (int positionCounter = 0; positionCounter < WIDTH; positionCounter++) {
    if (positionCounter < 8)
      tone(buzzer, notes[positionCounter]);
    else
      noTone(buzzer);
    lcd.scrollDisplayRight();
    delay(150);
  }
  delay(2000);
  lcd.clear();
}

void displayMenuScreen() {
  switch (menu) {
    case MENU_PLAY:
      lcd.clear();
      lcd.setCursor(0, 0);
      lcd.print("> Play");
      lcd.setCursor(0, 1);
      lcd.print("  Leaderboard");
      break;
    case MENU_LEADERBOARD:
      lcd.clear();
      lcd.setCursor(0, 0);
      lcd.print("  Play");
      lcd.setCursor(0, 1);
      lcd.print("> Leaderboard");
      break;
    default:
      break;      
  }
}

void prettyPrintTime(long int t) {
  if (t == MAX_TIME) {
    lcd.print("No entry");
    return;
  }
  lcd.print(t / 60);
  lcd.print(":");
  if (t % 60 < 10) {
    lcd.print("0");
  }
  lcd.print(t % 60);
}

void displayVictoryScreen(long int time_taken) {
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("CONGRATULATIONS!");
  lcd.setCursor(0, 1);
  lcd.print("Time: ");
  prettyPrintTime(time_taken);
  for (int note = 0; note < 16; note++) {
    tone(buzzer, victory_notes[note]);
    delay(150);
  }
  delay(300);
  noTone(buzzer);
}

void displayLeaderboardScreen() {
  switch (leaderboard_page) {
    case LDR_PG_1:
      lcd.clear();
      lcd.setCursor(0, 0);
      lcd.print("< Press");
      lcd.setCursor(0, 1);
      lcd.print("1. ");
      prettyPrintTime(leaderboard[0]);
      break;
    case LDR_PG_2:
      lcd.clear();
      lcd.setCursor(0, 0);
      lcd.print("2. ");
      prettyPrintTime(leaderboard[1]);
      lcd.setCursor(0, 1);
      lcd.print("3. ");
      prettyPrintTime(leaderboard[2]);
      break;
    default:
      break;
  }
}

void loadLeaderboard() {
  // open leaderboard file
  myFile = SD.open("example.txt", O_READ| O_WRITE | O_CREAT);

  if (myFile) {
    myFile.seek(0);
    for (int i = 0; i < LEADERBOARD_ENTRIES; i++) {
      if (myFile.available()) {
        leaderboard[i] = myFile.parseInt();
      } else {
        leaderboard[i] = MAX_TIME;
      }
    }
  }
  
  myFile.close();
}

void updateLeaderboard(long int new_time) {
  loadLeaderboard();
  
  // establish new top 3 times
  if (new_time < leaderboard[0]) {
    leaderboard[2] = leaderboard[1];
    leaderboard[1] = leaderboard[0];
    leaderboard[0] = new_time;
  } else if (new_time < leaderboard[1]) {
    leaderboard[2] = leaderboard[1];
    leaderboard[1] = new_time;
  } else if (new_time < leaderboard[2]) {
    leaderboard[2] = new_time;
  }

  // write new times
  myFile = SD.open("example.txt", O_READ| O_WRITE | O_CREAT);

  if (myFile) {
    myFile.seek(0);
    for (int i = 0; i < LEADERBOARD_ENTRIES; i++) {
      myFile.println(leaderboard[i]);
    }
  }
  
  myFile.close();
}

void initScreen() {
  // initialize screen buffer
  for (int row = 0; row < HEIGHT; row++)
    for (int pos = 0; pos < WIDTH; pos++)
      screen[row][pos] = NO_DISPLAY;

  // place the player's ship
  screen[HEIGHT - 1][ship.pos] = SHIP_DISPLAY;

  // generate aliens
  switch (level) {
    case 1:
      alien_no = 4;
      level_movement_timer = 6;
      level_direction_steps = 3;
      break;
    case 2:
      alien_no = 6;
      level_movement_timer = 5;
      level_direction_steps = 2;
      break;
    case 3:
      alien_no = 8;
      level_movement_timer = 4;
      level_direction_steps = 1;
      break;
    case 4:
      alien_no = 8;
      level_movement_timer = 3;
      level_direction_steps = 3;
      break;
    case 5:
      alien_no = 10;
      level_movement_timer = 2;
      level_direction_steps = 2;
      break;
    default:
      break;
  }

  alien_movement_direction = random(2);
  if (alien_movement_direction == 0)
    alien_movement_direction--;
  alien_movement_timer = level_movement_timer;
  alien_direction_steps = level_direction_steps;

  for (int pos = (WIDTH - alien_no) / 2 ; pos < (WIDTH + alien_no) / 2; pos++)
    screen[0][pos] = ALIEN_DISPLAY;
}

void moveAliensLeft() {
  for (int pos = 0; pos < WIDTH; pos++) {
    if (screen[0][pos + 1] == ALIEN_DISPLAY) {
      screen[0][pos] = ALIEN_DISPLAY;
      screen[0][pos + 1] = NO_DISPLAY;
    }
  }
}

void moveAliensRight() {
  for (int pos = WIDTH - 1; pos >= 0; pos--) {
    if (screen[0][pos - 1] == ALIEN_DISPLAY) {
      screen[0][pos] = ALIEN_DISPLAY;
      screen[0][pos - 1] = NO_DISPLAY;
    }
  }
}

bool checkLevelCompletion() {
  for (int pos = 0; pos < WIDTH; pos++)
    if (screen[0][pos] == ALIEN_DISPLAY)
      return false;
  return true;
}

void play() {
  // read the potentiometer on A0:
  int sensorReading = analogRead(A0);
  
  // map the result to 200 - 1000:
  int delayTime = map(sensorReading, 0, 1023, 200, 1000);

  // read data from joystick
  int x_data = analogRead(joystick_X);
  int btn_data = digitalRead(joystick_BTN);
  
  // handle ship movement
  if (x_data > 600 && ship.pos < WIDTH - 1) {
    screen[HEIGHT - 1][ship.pos] = NO_DISPLAY;
    ship.pos += 1;
    screen[HEIGHT - 1][ship.pos] = SHIP_DISPLAY;
  }

  if (x_data < 400 && ship.pos > 0) {
    screen[HEIGHT - 1][ship.pos] = NO_DISPLAY;
    ship.pos -= 1;
    screen[HEIGHT - 1][ship.pos] = SHIP_DISPLAY;
  }

  // handle ship shooting & button pressing
  if (ship.bullet_active && ship.bullet_row > 0) {
    screen[ship.bullet_row][ship.bullet_pos] = NO_DISPLAY;
    ship.bullet_row--;
    screen[ship.bullet_row][ship.bullet_pos] = BULLET_DISPLAY;
  } else if (ship.bullet_row == 0) {
    if (screen[ship.bullet_row][ship.bullet_pos] != ALIEN_DISPLAY)
      screen[ship.bullet_row][ship.bullet_pos] = NO_DISPLAY;
    ship.bullet_active = false;
    ship.bullet_pos = -1;
    ship.bullet_row = -1;
  }
  
  if (btn_data == 0 && !ship.bullet_active) {
    ship.bullet_active = true;
    ship.bullet_pos = ship.pos;
    ship.bullet_row = 2;
    screen[ship.bullet_row][ship.bullet_pos] = BULLET_DISPLAY;
    tone(buzzer, 400, 150);
  }

  // handle alien movement
  if (alien_movement_timer == 0) {
    if (alien_movement_direction == -1)
      moveAliensLeft();
    else
      moveAliensRight();
    alien_direction_steps--;
    alien_movement_timer = level_movement_timer;
  } else {
    alien_movement_timer--;  
  }

  if (alien_direction_steps == 0) {
    alien_movement_direction *= -1;
    alien_direction_steps = level_direction_steps;
  }

  // render shapes on screen
  updateScreen();
  delay(delayTime / 2);

  // next level
  if (checkLevelCompletion()) {
    level++;
    if (level > MAX_LEVEL) {
      end_time = millis();

      long int time_taken = (end_time - start_time) / 1000;

      displayVictoryScreen(time_taken);

      updateLeaderboard(time_taken);
      
      play_game = false;
    } else {
      displayLevelScreen(level);
      initScreen();
    }
  }
  
}

void setup() {
  // open serial communications and wait for port to open:
  Serial.begin(9600);
  while (!Serial) {
    ; // wait for serial port to connect. Needed for native USB port only
  }

  Serial.print("Initializing SD card...");

  if (!SD.begin(SD_module_CS)) {
    Serial.println("initialization failed!");
    while (1);
  }
  Serial.println("initialization done.");

  if (SD.exists("example.txt")) {
    Serial.println("example.txt exists.");
  } else {
    Serial.println("example.txt doesn't exist.");
  }

  Serial.println("Opening example.txt...");
  myFile = SD.open("example.txt", O_READ| O_WRITE | O_CREAT);
  myFile.close();

  if (SD.exists("example.txt")) {
    Serial.println("example.txt exists.");
  } else {
    Serial.println("example.txt doesn't exist.");
  }

  myFile = SD.open("example.txt", O_READ| O_WRITE | O_CREAT);
  myFile.seek(0);
  while (myFile.available()) {
    Serial.println(myFile.parseInt());
  }
  myFile.close();
  
  // initialize LCD and set up the number of columns and rows
  lcd.begin(WIDTH, HEIGHT / 2);

  // create shapes
  lcd.createChar(0, ship_shape);
  lcd.createChar(1, ship_bullet_shape);
  lcd.createChar(2, alien_shape);
  lcd.createChar(3, alien_bullet_shape);
  lcd.createChar(4, up_bullet_shape);
  lcd.createChar(5, down_bullet_shape);

  // joystick setup
  pinMode(joystick_X, INPUT);
  pinMode(joystick_Y, INPUT);
  pinMode(joystick_BTN, INPUT);
  digitalWrite(joystick_BTN, HIGH);

  // buzzer setup
  pinMode(buzzer, OUTPUT);

  // welcoming screen
  displayWelcomeScreen();
}

void loop() {
  // read data from joystick
  int y_data = analogRead(joystick_Y);
  int btn_data = digitalRead(joystick_BTN);

  if (!play_game && !see_leaderboard) {
    // display menu
    displayMenuScreen();

    // handle menu item selection
    if (y_data < 400 && menu == MENU_LEADERBOARD) {
      menu = MENU_PLAY;
      tone(buzzer, 330, 150);
    }
    else if (y_data > 600 && menu == MENU_PLAY) {
      menu = MENU_LEADERBOARD;
      tone(buzzer, 330, 150);
    }

    // handle button press
    if (btn_data == 0) { 
      if (menu == MENU_PLAY) {
        play_game = true;

        start_time = millis();
        
        level = 1;
  
        // first level screen
        displayLevelScreen(level);
  
        // initialize screen components (ship, enemies, timers)
        initScreen();
      }

      if (menu == MENU_LEADERBOARD) {
        loadLeaderboard();
        tone(buzzer, 520, 150);
        see_leaderboard = true;
      }
      delay(150);
    }
    
    delay(150);
  } else if (see_leaderboard) {
    // display leaderboard
    displayLeaderboardScreen();

    // handle leaderboard page selection
    if (y_data < 400 && leaderboard_page == LDR_PG_2) {
      leaderboard_page = LDR_PG_1;
      tone(buzzer, 330, 150);
    }
    else if (y_data > 600 && leaderboard_page == LDR_PG_1) {
      leaderboard_page = LDR_PG_2;
      tone(buzzer, 330, 150);
    }

    // handle button press on first page
    if (btn_data == 0 && leaderboard_page == LDR_PG_1) {
      see_leaderboard = false;
      tone(buzzer, 520, 150);
      delay(150);
    }

    delay(150);
  } else {
    play();
  }
}
